import setuptools

setuptools.setup(
  	name='rexpy',
    version='0.0.1',
    author='Nick De Breuck',
    author_email='nick.debreuck@outlook.com',
    description='Python framework used with Rex Engine',
    url='https://github.com/Dyronix/Rex',
    packages=setuptools.find_packages(),
    classifiers=[
      # https://pypi.org/classifiers/
      "Programming Language :: Python :: 3",
      "License :: OSI Approved :: MIT License",
      "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=["requests", "termcolor", "psutil", "filelock"],
)


import os
import time
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler, RegexMatchingEventHandler, FileSystemEvent, FileSystemMovedEvent

class DirWatcher():
    def __init__(self, dir : str, bRecursive : bool):
        self.event_handler = FileSystemEventHandler()

        self.event_handler.on_created = self._on_created
        self.event_handler.on_deleted = self._on_deleted
        self.event_handler.on_modified = self._on_modified
        self.event_handler.on_moved = self._on_moved
    
        self.observer = Observer()
        self.observer.schedule(self.event_handler, dir, recursive=bRecursive)

        self.created_or_modified_files : list[str] = []
        self.deleted_files : list[str] = []

    def __enter__(self):
        self.observer.start()       
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.observer.stop()
        self.observer.join()

    def _on_created(self, event : FileSystemEvent):
        full_path = os.path.normpath(os.path.join(os.getcwd(), event.src_path))
        if full_path not in self.created_or_modified_files:
            self.created_or_modified_files.append(full_path)
    
    def _on_deleted(self, event : FileSystemEvent):
        full_path = os.path.normpath(os.path.join(os.getcwd(), event.src_path))
        if full_path not in self.deleted_files:
            self.deleted_files.append(os.path.normpath(full_path))
    
    def _on_modified(self, event : FileSystemEvent):
        full_path = os.path.normpath(os.path.join(os.getcwd(), event.src_path))
        if full_path not in self.created_or_modified_files:
            self.created_or_modified_files.append(full_path)
    
    def _on_moved(self, event : FileSystemMovedEvent):
        full_dest_path = os.path.normpath(os.path.join(os.getcwd(), event.dest_path))
        if full_dest_path not in self.created_or_modified_files:
            self.created_or_modified_files.append(full_dest_path)
        
        full_src_path = os.path.normpath(os.path.join(os.getcwd(), event.src_path))
        if full_src_path not in self.deleted_files:
            self.deleted_files.append(full_src_path)

    def filter_created_or_modified_files(self, func):
        return list(filter(func, self.created_or_modified_files))
    
    def filter_deleted_files(self, func):
        return list(filter(func, self.deleted_files))




import os
from pathlib import Path
import required_tools
import util
import subprocess
import re
import diagnostics
import shutil

htlm_report_folder = "lcov"

def create_index_rawdata(rawdataPath):
  folder = Path(rawdataPath).parent
  output_path = os.path.join(folder, f"{Path(rawdataPath).stem}.profdata")
  llvm_profdata_path = required_tools.tool_paths_dict["llvm_profdata_path"]
  os.system(f"{llvm_profdata_path} merge -sparse {rawdataPath} -o {output_path}")

  return output_path

def get_line_oriented_report_filename(profDataPath):
  return os.path.join(Path(profDataPath).parent, f"{Path(profDataPath).stem}_line.html")

def get_file_level_summary_filename(profDataPath):
  return os.path.join(Path(profDataPath).parent, f"{Path(profDataPath).stem}_file.report")

def get_lcov_filename(profDataPath):
  return os.path.join(Path(profDataPath).parent, f"{Path(profDataPath).stem}_lcov.info")

def get_lcov_unmangled_filename(profDataPath):
  return os.path.join(Path(profDataPath).parent, f"{Path(profDataPath).stem}_lcov_unmangled.info")

def create_line_oriented_report(programPath, profDataPath):
  llvm_cov_path = required_tools.tool_paths_dict["llvm_cov_path"]
  log_file_path = get_line_oriented_report_filename(profDataPath)
  if os.path.exists(log_file_path):
    os.remove(log_file_path)

  f = open(log_file_path, "w")
  cmd = f"{llvm_cov_path} show -format=html {programPath} -instr-profile={profDataPath} >> {log_file_path}"
  f.write(f"# This file was generated by running the following command:\n")
  f.write(f"# {cmd}\n")
  f.close()
  os.system(cmd) # using >> is a hack to get the logs into a file, capturing stdout lines crashes llvm
  return log_file_path
  
def create_file_level_summary(programPath, profDataPath):
  llvm_cov_path = required_tools.tool_paths_dict["llvm_cov_path"]
  log_file_path = get_file_level_summary_filename(profDataPath)
  if os.path.exists(log_file_path):
    os.remove(log_file_path)

  f = open(log_file_path, "w")
  cmd = f"{llvm_cov_path} report {programPath} -instr-profile={profDataPath} >> {log_file_path}"
  f.write(f"# This file was generated by running the following command:\n")
  f.write(f"# {cmd}\n")
  f.close()
  os.system(cmd) # using >> is a hack to get the logs into a file, capturing stdout lines crashes llvm
  return log_file_path

def __create_mangled_lcov_info(programPath, profDataPath):
  llvm_cov_path = required_tools.tool_paths_dict["llvm_cov_path"]
  log_file_path = get_lcov_filename(profDataPath)
  cmd = f"{llvm_cov_path} export -format=lcov {programPath} -instr-profile={profDataPath} >> {log_file_path}"
  os.system(cmd)
  return log_file_path

def __unmangle_function_names(logFilePath, profDataPath):
  # the lcov info file will have mangled names, unfortunately using -Xdemangler doesn't work so we have to manually parse by undname (on Windows)
  # this works, but will result in a bit of an invalid .info file as it parsed ',' to get to the next value, and templated functions will use ',' tokens to separate arguments
  # so we'll parse these .info files ourselves after they're unmangled, changing all ',' to a different token to have a proper html report

  #  0x0001  Remove leading underscores from Microsoft extended keywords
  #  0x0002  Disable expansion of Microsoft extended keywords
  #  0x0004  Disable expansion of return type for primary declaration
  #  0x0008  Disable expansion of the declaration model
  #  0x0010  Disable expansion of the declaration language specifier
  #  0x0060  Disable all modifiers on the 'this' type
  #  0x0080  Disable expansion of access specifiers for members
  #  0x0100  Disable expansion of 'throw-signatures' for functionsand pointers to
  #  functions
  #  0x0200  Disable expansion of 'static' or 'virtual'ness of members
  #  0x0400  Disable expansion of Microsoft model for UDT returns
  #  0x0800  Undecorate 32 - bit decorated names
  #  0x1000  Crack only the name for primary declaration
  #  return just[scope::]name.Does expand template params
  #  0x2000  Input is just a type encoding; compose an abstract declarator
  #  0x8000  Disable enum / class / struct / union prefix
  #  0x20000 Disable expansion of __ptr64 keyword

  # creating the unmangled .info file
  undname_path = required_tools.tool_paths_dict["undname_path"]
  flags : int = 0x0001 | 0x0002 | 0x0080 | 0x8000
  unmangled_log_file_path = get_lcov_unmangled_filename(profDataPath)
  cmd = f"\"{undname_path}\" {flags} {logFilePath} > {unmangled_log_file_path}"
  os.system(cmd)

  # now parse it and change the templated tokens
  f = open(unmangled_log_file_path, "r+")
  lines = f.readlines()
  f.seek(0)

  new_lines : list[str] = []
  for line in lines:
    if "FN:" not in line and "FNDA:" not in line: #function names start with FN, function execution counts with FNDA
      new_lines.append(line)
      continue

    # if it's found, then the next element is the function name
    first_comma = line.find(',')
    func_name = line[first_comma+1:]
    new_func_name = func_name.replace(',', '|')
    if new_func_name != func_name:
      line = line.replace(func_name, new_func_name)
    new_lines.append(line)

  for line in new_lines:
    f.write(line)
  f.truncate()
  f.close()

  return unmangled_log_file_path

def __generate_html_reports(unmangledLogFilePath):
  lcov_path = required_tools.tool_paths_dict["lcov_path"]
  cmd = f"perl {lcov_path} {unmangledLogFilePath} -q -o {os.path.join(Path(unmangledLogFilePath).parent, htlm_report_folder)}"
  os.system(cmd)

def create_lcov_report(programPath, profDataPath):
  if os.path.exists(htlm_report_folder):
    shutil.rmtree(htlm_report_folder)
  
  log_file_path = __create_mangled_lcov_info(programPath, profDataPath)
  unmangled_log_file_path = __unmangle_function_names(log_file_path, profDataPath)
  __generate_html_reports(unmangled_log_file_path)

class CoverageCategory:
  def __init__(self, total, missed, cover):
    self._total : str = total   # total number of elements of the category
    self._missed : str = missed # missed number of elements of the category
    self._cover : str = cover  # percent of the category covered

  def coverage(self):
    if self._cover == '-': # no elements of category in file
      return 100
    else:
      return float(self._cover.strip('%'))

  def total_str(self):
    return f"total: {self._total}"

  def missed_str(self):
    return f"missed: {self._total}"

  def covered_str(self):
    return f"covered: {self._total}"

class FileSummary:
  def __init__(self, line : str):
    words = line.split()
    self._filename = words[0]
    self._regions_summary = CoverageCategory(words[1], words[2], words[3])    
    self._functions_summary = CoverageCategory(words[4], words[5], words[6])    
    self._lines_summary = CoverageCategory(words[7], words[8], words[9])    
    self._branches_summary = CoverageCategory(words[10], words[11], words[12])    

  def to_string(self):
    result = ""
    result += f"#################################"
    result += f"coverage for file: {self._filename}"
    result += f"REGIONS:"
    result += f"---------------------------------"
    result += self._regions_summary.total_str()
    result += self._regions_summary.missed_str()
    result += self._regions_summary.covered_str()
    result += f"FUNCTIONS:"
    result += f"---------------------------------"
    result += self._functions_summary.total_str()
    result += self._functions_summary.missed_str()
    result += self._functions_summary.covered_str()
    result += f"LINES:"
    result += f"---------------------------------"
    result += self._lines_summary.total_str()
    result += self._lines_summary.missed_str()
    result += self._lines_summary.covered_str()
    result += f"BRANCHES:"
    result += f"---------------------------------"
    result += self._branches_summary.total_str()
    result += self._branches_summary.missed_str()
    result += self._branches_summary.covered_str()

    return result

  def filename(self):
    return self._filename

  def coverage(self):
    res = 0
    res += self._regions_summary.coverage()
    res += self._functions_summary.coverage()
    res += self._lines_summary.coverage()
    res += self._branches_summary.coverage()
    return res / 4

def parse_file_summary(filepath):
  file = open(filepath, "r")

  # format of file summary file:
  # lines starting with '#' are comments
  # lines with only consisting of '-' are separators 
  
  # Column names:
  # Filename | Regions | Missed Regions | Cover | Functions | Missed Functions | Executed | Lines | Missed Lines | Cover | Branches | Missed Branches | Cover
  # <separator>
  # file 1
  # file 2
  # ...
  # <separator>
  # Total

  file_summaries : list[FileSummary] = []
  total_summary : FileSummary
  lines = file.readlines()
  column_names_processed = False
  for line in lines:
    if line.startswith('#'):
      continue

    if re.search(r'[^-]', line).start() >= 200: # report files use a lot of '-' tokens as separators
      column_names_processed = True
      continue

    # the column names are displayed first, only after that do the actual files get shown
    if column_names_processed == False:
      continue

    line_words = line.split()
    if len(line_words) == 0:
      continue

    if "Files which contain no functions" in line:
      continue

    if line_words[0] == "TOTAL":  # filename == TOTAL
      total_summary = FileSummary(line)
    else:
      file_summaries.append(FileSummary(line))

  result = 0

  for file_summary in file_summaries:
    if file_summary.coverage() != 100:
      result = 1
      # diagnostics.log_err(f"File {file_summary.filename()} was not fully covered, please see below for more details")
      # diagnostics.log_err(f"Alternatively, investigate the line coverage report file for this file")    
      # diagnostics.log_err(f"This file is located at {Path(filepath).parent}")    

  if result != 0:
    diagnostics.log_err(f"No full coverage. More info found: {filepath}")
    diagnostics.log_err(f"Alternatively, investigate the line coverage file or html report located at {Path(filepath).parent}")    

  return result



import os
import regis.required_tools
import regis.rex_json
import regis.util
import regis.diagnostics
import regis.subproc

from pathlib import Path

from requests.structures import CaseInsensitiveDict

tool_paths_dict = regis.required_tools.tool_paths_dict

class NinjaProject:
  def __init__(self, filepath):
    self.json_blob : dict = regis.rex_json.load_file(filepath)
    self.filepath = filepath
    self.project_name = list(self.json_blob.keys())[0].lower() # the project name is the root key

  def ninja_file(self, compiler : str, config : str, buildDependencies : bool):
    if buildDependencies:
      return self.json_blob[self.project_name][compiler.lower()][config.lower()]["ninja_file"]
    else:
      return self.json_blob[self.project_name][compiler.lower()][config.lower()]["ninja_file_no_deps"]

  def dependencies(self, compiler : str, config : str):
    return self.json_blob[self.project_name.lower()][compiler.lower()][config.lower()]["dependencies"]

  def clean(self, compiler : str, config : str, buildDependencies : bool):
    ninja_path = tool_paths_dict["ninja_path"]
    regis.diagnostics.log_info(f'Cleaning intermediates')
    
    r = self._valid_args_check(compiler, config)
    if r != 0:
      return r

    proc = regis.subproc.run(f"{ninja_path} -f {self.ninja_file(compiler, config, buildDependencies)} -t clean")
    proc.wait()

    r = proc.returncode
    return r
  
  def build(self, compiler : str, config : str, buildDependencies : bool):
    r = 0

    r = self._valid_args_check(compiler, config)
    if r != 0:
      return r

    # make sure to build the dependencies first
    if buildDependencies:
      r |= self._build_dependencies(compiler, config)

    regis.diagnostics.log_info(f"Building: {self.project_name} - {config} - {compiler}")

    # then build the project we specified
    ninja_path = tool_paths_dict["ninja_path"]
    cmd = f"{ninja_path} -f {self.ninja_file(compiler, config, buildDependencies)}"
    regis.diagnostics.log_info(f'executing: {cmd}')
    proc = regis.subproc.run(cmd)
    proc.wait()
    r = proc.returncode

    # show error if any build failed
    if r != 0:
      regis.diagnostics.log_err(f"Failed to build {self.project_name}")

    return r
  
  def rebuild(self, compiler : str, config : str, buildDependencies : bool):
    r = self.clean(compiler, config, buildDependencies)
    if r != 0:
      return r
    
    r = self.build(compiler, config, buildDependencies)

    return r
   
  def _valid_args_check(self, compiler : str, config : str):
    if compiler not in self.json_blob[self.project_name]:
      regis.diagnostics.log_err(f"no compiler '{compiler}' found for project '{self.project_name}'")
      return 1
  
    if config not in self.json_blob[self.project_name][compiler]:
      regis.diagnostics.log_err(f"error in {self.filepath}")
      regis.diagnostics.log_err(f"no config '{config}' found in project '{self.project_name}' for compiler '{compiler}'")
      return 1
    
    return 0

  def _build_dependencies(self, compiler, config):
    dependencies = self.json_blob[self.project_name][compiler][config]["dependencies"]

    r = 0
    for dependency in dependencies:      
      dependency_project_name = Path(dependency).stem
      regis.diagnostics.log_info(f'Building dependency: {self.project_name} -> {dependency_project_name}')

      dependency_project = NinjaProject(dependency)
      r |= dependency_project.build(compiler, config, buildDependencies=True)

    return r

def find_sln(directory):
  dirs = os.listdir(directory)

  res = []

  for dir in dirs:
    full_path = os.path.join(directory, dir)
    if os.path.isfile(full_path) and Path(full_path).suffix == ".nsln":
      res.append(full_path)
    
  return res

def __launch_new_build(sln_file : str, projectName : str, config : str, compiler : str, shouldBuild : bool, shouldClean : bool, buildDependencies = False):
  sln_jsob_blob = CaseInsensitiveDict(regis.rex_json.load_file(sln_file))
  
  if projectName not in sln_jsob_blob:
    regis.diagnostics.log_err(f"project '{projectName}' was not found in solution, have you generated it?")
    return 1
  
  project_file_path = sln_jsob_blob[projectName]    
  project = NinjaProject(project_file_path)

  compiler_lower = compiler.lower()
  config_lower = config.lower()

  r = 0

  if shouldClean:
    r |= project.clean(compiler_lower, config_lower, buildDependencies)

  if shouldBuild:
    r |= project.build(compiler_lower, config_lower, buildDependencies)

  return r

def __look_for_sln_file_to_use(slnFile : str):
  if slnFile == "":
    root = regis.util.find_root()
    sln_files = find_sln(root)

    if len(sln_files) > 1:
      regis.diagnostics.log_err(f'more than 1 nsln file was found in the cwd, please specify which one you want to use')
    
      for file in sln_files:
        regis.diagnostics.log_err(f'-{file}')
    
      return ""
    
    if len(sln_files) == 0:
      regis.diagnostics.log_err(f'no nlsn found in {root}')
      return ""

    slnFile = sln_files[0]
  elif not os.path.exists(slnFile):
    regis.diagnostics.log_err(f'solution path {slnFile} does not exist')
    return ""
  
  return slnFile

def new_build(project : str, config : str, compiler : str, shouldBuild : bool = False, shouldClean : bool = False, slnFile : str = "", buildDependencies : bool = False):
  slnFile = __look_for_sln_file_to_use(slnFile)

  if slnFile == "":
    regis.diagnostics.log_err("unable to find solution, aborting..")
    return 1
  
  res = __launch_new_build(slnFile, project, config, compiler, shouldBuild, shouldClean, buildDependencies)
  return res
  


#!/usr/bin/env python3
#
#===- run-clang-tidy.py - Parallel clang-tidy runner --------*- python -*--===#
#
# Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
# See https://llvm.org/LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
#===-----------------------------------------------------------------------===#
# FIXME: Integrate with clang-tidy-diff.py


"""
Parallel clang-tidy runner
==========================

Runs clang-tidy over all files in a compilation database. Requires clang-tidy
and clang-apply-replacements in $PATH.

Example invocations.
- Run clang-tidy on all files in the current working directory with a default
  set of checks and show warnings in the cpp files and all project headers.
    run-clang-tidy.py $PWD

- Fix all header guards.
    run-clang-tidy.py -fix -checks=-*,llvm-header-guard

- Fix all header guards included from clang-tidy and header guards
  for clang-tidy headers.
    run-clang-tidy.py -fix -checks=-*,llvm-header-guard extra/clang-tidy \
                      -header-filter=extra/clang-tidy

Compilation database setup:
http://clang.llvm.org/docs/HowToSetupToolingForLLVM.html
"""

from __future__ import print_function

import argparse
import glob
import json
import multiprocessing
import os
import queue
import re
import shutil
import subprocess
import sys
import tempfile
import threading
import traceback
import time

import regis.util
import regis.diagnostics
import regis.rex_json

try:
  import yaml
except ImportError:
  yaml = None

processed_files : dict[str, float] = {}
active_pids = []

def strtobool(val):
  """Convert a string representation of truth to a bool following LLVM's CLI argument parsing."""

  val = val.lower()
  if val in ['', 'true', '1']:
    return True
  elif val in ['false', '0']:
    return False

  # Return ArgumentTypeError so that argparse does not substitute its own error message
  raise argparse.ArgumentTypeError(
    "'{}' is invalid value for boolean argument! Try 0 or 1.".format(val)
  )


def find_compilation_database(path):
  """Adjusts the directory until a compilation database is found."""
  result = os.path.realpath('./')
  while not os.path.isfile(os.path.join(result, path)):
    parent = os.path.dirname(result)
    if result == parent:
      print('Error: could not find compilation database.')
      sys.exit(1)
    result = parent
  return result


def make_absolute(f, directory):
  if os.path.isabs(f):
    return f
  return os.path.normpath(os.path.join(directory, f))


def get_tidy_invocation(f, clang_tidy_binary, checks, tmpdir, build_path,
                        header_filter, allow_enabling_alpha_checkers,
                        extra_arg, extra_arg_before, quiet, config_file_path,
                        config, line_filter, use_color, plugins):
  """Gets a command line for clang-tidy."""
  start = [clang_tidy_binary]
  if allow_enabling_alpha_checkers:
    start.append('-allow-enabling-analyzer-alpha-checkers')
  if header_filter is not None:
    start.append('-header-filter=' + header_filter)
  if line_filter is not None:
    start.append('-line-filter=' + line_filter)
  if use_color is not None:
    if use_color:
      start.append('--use-color')
    else:
      start.append('--use-color=false')
  if checks:
    start.append('-checks=' + checks)
  if tmpdir is not None:
    start.append('-export-fixes')
    # Get a temporary file. We immediately close the handle so clang-tidy can
    # overwrite it.
    (handle, name) = tempfile.mkstemp(suffix='.yaml', dir=tmpdir)
    os.close(handle)
    start.append(name)
  for arg in extra_arg:
      start.append('-extra-arg=%s' % arg)
  for arg in extra_arg_before:
      start.append('-extra-arg-before=%s' % arg)
  start.append('-p=' + build_path)
  if quiet:
      start.append('-quiet')
  if config_file_path:
      start.append('--config-file=' + config_file_path)
  elif config:
      start.append('-config=' + config)
  for plugin in plugins:
      start.append('-load=' + plugin)
  start.append(f)
  return start


def merge_replacement_files(tmpdir, mergefile):
  """Merge all replacement files in a directory into a single file"""
  # The fixes suggested by clang-tidy >= 4.0.0 are given under
  # the top level key 'Diagnostics' in the output yaml files
  mergekey = "Diagnostics"
  merged=[]
  for replacefile in glob.iglob(os.path.join(tmpdir, '*.yaml')):
    content = yaml.safe_load(open(replacefile, 'r'))
    if not content:
      continue # Skip empty files.
    merged.extend(content.get(mergekey, []))

  if merged:
    # MainSourceFile: The key is required by the definition inside
    # include/clang/Tooling/ReplacementsYaml.h, but the value
    # is actually never used inside clang-apply-replacements,
    # so we set it to '' here.
    output = {'MainSourceFile': '', mergekey: merged}
    with open(mergefile, 'w') as out:
      yaml.safe_dump(output, out)
  else:
    # Empty the file:
    open(mergefile, 'w').close()


def find_binary(arg, name, build_path):
  """Get the path for a binary or exit"""
  if arg:
    if shutil.which(arg):
      return arg
    else:
      raise SystemExit(
        "error: passed binary '{}' was not found or is not executable"
        .format(arg))

  built_path = os.path.join(build_path, "bin", name)
  binary = shutil.which(name) or shutil.which(built_path)
  if binary:
    return binary
  else:
    raise SystemExit(
      "error: failed to find {} in $PATH or at {}"
      .format(name, built_path))

def processed_filepath(build_path : str):
  return os.path.join(build_path, 'processed.json')

def should_process_file(file : str, build_path : str, config_path : str):
  global processed_files
  invocation_str = invocation_string(build_path, file, config_path)
  if invocation_str in processed_files:
    was_successful = processed_files[invocation_str]['successful']
    file_process_time = float(processed_files[invocation_str]['timestamp'])
    file_mod_time = os.path.getmtime(file)

    return not was_successful or file_mod_time > file_process_time
  
  return True
  
def apply_fixes(args, clang_apply_replacements_binary, tmpdir):
  """Calls clang-apply-fixes on a given directory."""
  invocation = [clang_apply_replacements_binary]
  invocation.append(' -ignore-insert-conflict')
  if args.format:
    invocation.append(' -format')
  if args.style:
    invocation.append(' -style=' + args.style)
  invocation.append(f' {tmpdir}')
  
  cmd = ''.join(invocation)
  proc = regis.util.run_subprocess(''.join(cmd))
  regis.util.wait_for_process(proc)
  
def invocation_string(build_path : str, file : str, config_path : str):
  return build_path + ' - ' + config_path + ' - -' + (file)

def run_tidy(args, clang_tidy_binary, tmpdir, build_path, queue, lock,
             failed_files):
  """Takes filenames out of queue and runs clang-tidy on them."""
  while True:
    name = queue.get()
    invocation = get_tidy_invocation(name, clang_tidy_binary, args.checks,
                                     tmpdir, build_path, args.header_filter,
                                     args.allow_enabling_alpha_checkers,
                                     args.extra_arg, args.extra_arg_before,
                                     args.quiet, args.config_file, args.config,
                                     args.line_filter, args.use_color,
                                     args.plugins)

    try:
      proc = subprocess.Popen(invocation, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
      active_pids.append(proc.pid)
      output, err = proc.communicate()
      active_pids.remove(proc.pid)
      if proc.returncode != 0:
        if proc.returncode < 0:
          msg = "%s: terminated by signal %d\n" % (name, -proc.returncode)
          err += msg.encode('utf-8')
        failed_files.append(name)
      with lock:
        sys.stdout.write(' '.join(invocation) + '\n' + output.decode('utf-8'))
        if len(err) > 0:
          sys.stdout.flush()
          sys.stderr.write(err.decode('utf-8'))
        invocation_str = invocation_string(build_path, name, args.config_file)
        processed_files[invocation_str] = {}
        processed_files[invocation_str]['timestamp'] = time.time()
        processed_files[invocation_str]['successful'] = proc.returncode == 0 # warnings are > 0, errors are < 0
    except Exception as Ex:
      regis.diagnostics.log_err(f'exception occurred: {Ex}')
    queue.task_done()



def run():
  parser = argparse.ArgumentParser(description='Runs clang-tidy over all files '
                                   'in a compilation database. Requires '
                                   'clang-tidy and clang-apply-replacements in '
                                   '$PATH or in your build directory.')
  parser.add_argument('-allow-enabling-alpha-checkers',
                      action='store_true', help='allow alpha checkers from '
                                                'clang-analyzer.')
  parser.add_argument('-clang-tidy-binary', metavar='PATH',
                      help='path to clang-tidy binary')
  parser.add_argument('-clang-apply-replacements-binary', metavar='PATH',
                      help='path to clang-apply-replacements binary')
  parser.add_argument('-checks', default=None,
                      help='checks filter, when not specified, use clang-tidy '
                      'default')
  config_group = parser.add_mutually_exclusive_group()
  config_group.add_argument('-config', default=None,
                      help='Specifies a configuration in YAML/JSON format: '
                      '  -config="{Checks: \'*\', '
                      '                       CheckOptions: {x: y}}" '
                      'When the value is empty, clang-tidy will '
                      'attempt to find a file named .clang-tidy for '
                      'each source file in its parent directories.')
  config_group.add_argument('-config-file', default=None,
                      help='Specify the path of .clang-tidy or custom config '
                      'file: e.g. -config-file=/some/path/myTidyConfigFile. '
                      'This option internally works exactly the same way as '
                      '-config option after reading specified config file. '
                      'Use either -config-file or -config, not both.')
  parser.add_argument('-header-filter', default=None,
                      help='regular expression matching the names of the '
                      'headers to output diagnostics from. Diagnostics from '
                      'the main file of each translation unit are always '
                      'displayed.')
  parser.add_argument('-line-filter', default=None,
                      help='List of files with line ranges to filter the'
                      'warnings.')
  if yaml:
    parser.add_argument('-export-fixes', metavar='filename', dest='export_fixes',
                        help='Create a yaml file to store suggested fixes in, '
                        'which can be applied with clang-apply-replacements.')
  parser.add_argument('-j', type=int, default=0,
                      help='number of tidy instances to be run in parallel.')
  parser.add_argument('files', nargs='*', default=['.*'],
                      help='files to be processed (regex on path)')
  parser.add_argument('-fix', action='store_true', help='apply fix-its')
  parser.add_argument('-format', action='store_true', help='Reformat code '
                      'after applying fixes')
  parser.add_argument('-style', default='file', help='The style of reformat '
                      'code after applying fixes')
  parser.add_argument('-use-color', type=strtobool, nargs='?', const=True,
                      help='Use colors in diagnostics, overriding clang-tidy\'s'
                      ' default behavior. This option overrides the \'UseColor'
                      '\' option in .clang-tidy file, if any.')
  parser.add_argument('-p', dest='build_path',
                      help='Path used to read a compile command database.')
  parser.add_argument('-extra-arg', dest='extra_arg',
                      action='append', default=[],
                      help='Additional argument to append to the compiler '
                      'command line.')
  parser.add_argument('-extra-arg-before', dest='extra_arg_before',
                      action='append', default=[],
                      help='Additional argument to prepend to the compiler '
                      'command line.')
  parser.add_argument('-quiet', action='store_true',
                      help='Run clang-tidy in quiet mode')
  parser.add_argument('-load', dest='plugins',
                      action='append', default=[],
                      help='Load the specified plugin in clang-tidy.')
  parser.add_argument('-incremental', action='store_true', default=False, help='run incrementally, skip files already processed run and haven\'t changed since last run')

  args = parser.parse_args()

  db_path = 'compile_commands.json'

  if args.build_path is not None:
    build_path = args.build_path
  else:
    # Find our database
    build_path = find_compilation_database(db_path)

  if args.incremental:
    regis.diagnostics.log_info(f'Running incremental mode')
    global processed_files
    processed_path = processed_filepath(build_path)
    if os.path.exists(processed_path):
      processed_files = regis.rex_json.load_file(processed_path)
    else:
      processed_files = {}

  clang_tidy_binary = find_binary(args.clang_tidy_binary, "clang-tidy",
                                  build_path)

  tmpdir = None
  if args.fix or (yaml and args.export_fixes):
    clang_apply_replacements_binary = find_binary(
      args.clang_apply_replacements_binary, "clang-apply-replacements",
      build_path)
    tmpdir = tempfile.mkdtemp()

  try:
    invocation = get_tidy_invocation("", clang_tidy_binary, args.checks,
                                     None, build_path, args.header_filter,
                                     args.allow_enabling_alpha_checkers,
                                     args.extra_arg, args.extra_arg_before,
                                     args.quiet, args.config_file, args.config,
                                     args.line_filter, args.use_color,
                                     args.plugins)
    invocation.append('-list-checks')
    invocation.append('-')
    if args.quiet:
      # Even with -quiet we still want to check if we can call clang-tidy.
      with open(os.devnull, 'w') as dev_null:
        subprocess.check_call(invocation, stdout=dev_null)
    else:
      subprocess.check_call(invocation)
  except:
    print("Unable to run clang-tidy.", file=sys.stderr)
    sys.exit(1)

  # Load the database and extract all files.
  database = json.load(open(os.path.join(build_path, db_path)))
  files = set([make_absolute(entry['file'], entry['directory'])
           for entry in database])

  max_task = args.j
  if max_task == 0:
    max_task = multiprocessing.cpu_count()

  # Build up a big regexy filter from all command line arguments.
  file_name_re = re.compile('|'.join(args.files))

  return_code = 0
  try:
    # Spin up a bunch of tidy-launching threads.
    task_queue = queue.Queue(max_task)
    # List of files with a non-zero return code.
    failed_files = []
    lock = threading.Lock()
    for _ in range(max_task):
      t = threading.Thread(target=run_tidy,
                           args=(args, clang_tidy_binary, tmpdir, build_path,
                                 task_queue, lock, failed_files))
      t.daemon = True
      t.start()

    # Fill the queue with files.
    num_files_skipped = 0
    for name in files:
      if file_name_re.search(name):
        if not args.incremental or should_process_file(name, build_path, args.config_file):
          task_queue.put(name)
        else:
          num_files_skipped += 1

    # Wait for all threads to be done.
    task_queue.join()

    regis.diagnostics.log_info(f'skipped {num_files_skipped} files')

    if len(failed_files):
      return_code = 1

  except KeyboardInterrupt:
    # This is a sad hack. Unfortunately subprocess goes
    # bonkers with ctrl-c and we start forking merrily.
    print('\nCtrl-C detected, goodbye.')
    if tmpdir:
      shutil.rmtree(tmpdir)

    for pid in active_pids:
      try:
        os.kill(pid, 9)
      except Exception as ex:
        # this is a hack on Windows which can sometimes fail to kill a process
        try:
          subprocess.check_output("Taskkill /PID %d /F" % pid)
        except Exception as ex:
          pass

  if yaml and args.export_fixes:
    print('Writing fixes to ' + args.export_fixes + ' ...')
    try:
      merge_replacement_files(tmpdir, args.export_fixes)
    except:
      print('Error exporting fixes.\n', file=sys.stderr)
      traceback.print_exc()
      return_code=1

  if args.fix:
    print('Applying fixes ...')
    try:
      apply_fixes(args, clang_apply_replacements_binary, tmpdir)
    except:
      print('Error applying fixes.\n', file=sys.stderr)
      traceback.print_exc()
      return_code = 1

  if tmpdir:
    shutil.rmtree(tmpdir)
    
  regis.rex_json.save_file(processed_filepath(build_path), processed_files)

  sys.exit(return_code)


if __name__ == '__main__':
  run()


import os

import regis.util
import regis.rex_json
import regis.diagnostics

changes_cache_filepath = os.path.join(".git", "file_changed")

def __zsplit(s: str) -> list[str]:
  s = s.strip('\0')
  s = s.strip('\n')
  if s:
      return s.split('\n')
  else:
      return []
  
def get_staged_files():
  cmd = 'git diff --staged --name-only --no-ext-diff'
  output, errc = regis.util.run_and_get_output(cmd)
  return __zsplit(output)

def get_unstaged_files():
  cmd = 'git diff --name-only --no-ext-diff'
  output, errc = regis.util.run_and_get_output(cmd)
  return __zsplit(output)

def get_local_branchname():
  cmd = 'git rev-parse --abbrev-ref HEAD'
  output, errc = regis.util.run_and_get_output(cmd)
  return output.strip('\n')

def cache_commit_changes(branch : str, files : list[str]):
  full_changes_cache_filepath = os.path.join(regis.util.find_root(), changes_cache_filepath)

  cached_changes = {}
  if os.path.exists(full_changes_cache_filepath):
    cached_changes = regis.rex_json.load_file(full_changes_cache_filepath)
  
  if not branch in cached_changes:
    cached_changes[branch] = []

  changes_in_branch = cached_changes[branch]

  for file in files:
    if file not in changes_in_branch:
      changes_in_branch.append(file)
      print(f"adding file")

  regis.rex_json.save_file(changes_cache_filepath, cached_changes)   

def get_cached_changes(branch):
  full_changes_cache_filepath = os.path.join(regis.util.find_root(), changes_cache_filepath)

  if os.path.exists(full_changes_cache_filepath):
    cached_changes = regis.rex_json.load_file(full_changes_cache_filepath)
    local_branch = branch
    
    if local_branch in cached_changes:
      return cached_changes[local_branch]
    
  return []

def stash_uncommitted_changes(stashName):
  cmd = f"git stash save {stashName} -k"
  output, errc = regis.util.run_and_get_output(cmd)
  regis.diagnostics.log_no_color(output)
  return output

def unstash(stashName):
  stash_idx, stash_branch, stash_name = find_stash(stashName)
  cmd = f"git stash apply {stash_idx}"
  output, errc = regis.util.run_and_get_output(cmd)
  regis.diagnostics.log_no_color(output)
  return output

def get_stash_list():
  cmd = f"git stash list"
  output, errc = regis.util.run_and_get_output(cmd)
  return __zsplit(output)

def find_stash(stashName):
  stashes = get_stash_list()
  branch_name = get_local_branchname()
  
  # stash format: stash@{0}: On devops/githooks: pre-push
  for stash in stashes:
    colon_idx = stash.find(':')
    stash_idx = stash[0: colon_idx]
    num_to_skip = len(": On ")
    branch_start = colon_idx + num_to_skip
    colon_idx = stash.find(':', colon_idx + 1)
    stash_branch = stash[branch_start: colon_idx]
    num_to_skip = len(": ")
    name_start = colon_idx + num_to_skip
    stash_name = stash[name_start: len(stash)]

    if (stashName.lower() == stash_name.lower()):
      return stash_idx, stash_branch, stash_name
    
  return None, None, None

def remove_stash(stashName):
  stash_idx, stash_branch, stash_name = find_stash(stashName)
  cmd = f"git stash drop {stash_idx}"
  output, errc = regis.util.run_and_get_output(cmd)
  regis.diagnostics.log_no_color(output)



import logging
import termcolor

class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    RESET = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

class Logger(object):
  def __init__(self, name = "logger", level = logging.DEBUG):
    if __import__("platform").system() == "Windows": # if running on Windows, enable console VT mode
      kernel32 = __import__("ctypes").windll.kernel32
      kernel32.SetConsoleMode(kernel32.GetStdHandle(-11), 7)
      del kernel32

    self.logger = logging.getLogger(name)
    self.logger.setLevel(level)

  def no_color(self, msg):
    print(msg)

  def debug(self, msg):
    termcolor.cprint(f"{msg}", "magenta")

  def info(self, msg):
    termcolor.cprint(f"{msg}", "light_green")

  def warning(self, msg):
    termcolor.cprint(f"{msg}", "yellow")

  def error(self, msg):
    termcolor.cprint(f"{msg}", "light_red")

  def critical(self, msg):
    self.logger.error(f"-- Critical Error Occurred --")
    self.logger.error(msg)
    raise Exception(msg)

class StreamLogger(Logger):
  def __init__(self, name = "logger", level = logging.DEBUG):
    super().__init__(name, level)

    sh = logging.StreamHandler()
    self.logger.addHandler(sh)

class FileLogger(Logger):
  def __init__(self, name = "logger", level = logging.DEBUG):
    super().__init__(name, level)

    fh = logging.FileHandler(f"{name}.log", "w")
    self.logger.addHandler(fh)

def logging_level_from_string(level):
  if level == "debug":
    return logging.DEBUG
  elif level == "info":
    return logging.INFO
  elif level == "warning":
    return logging.WARNING
  elif level == "error":
    return logging.ERROR
  elif level == "critical":
    return logging.CRITICAL

  return logging.NOTSET

__cout = StreamLogger("cout", logging.INFO)

def log_no_color(msg):
  __cout.no_color(f"{msg}")

def log_debug(msg):
  __cout.debug(f"{msg}")

def log_info(msg):
  __cout.info(f"{msg}")

def log_warn(msg):
  __cout.warning(f"{msg}")

def log_err(msg):
  __cout.error(f"{msg}")

def log_critical(msg):
  __cout.critical(f"{msg}")



import os
import subprocess
import shutil
from pathlib import Path
import regis.diagnostics
import regis.rex_json

settingsPathFromRoot = os.path.join("_build", "config", "settings.json")

def create_version_file(directory : str, tag : str):
    version = {
        "tag": tag
    }

    if not os.path.exists(directory):
      os.mkdir(directory)
      
    path = os.path.join(directory, "version.json")
    regis.rex_json.save_file(path, version)

def load_version_file(directory):
  version_file = os.path.join(directory, "version.json")
  if os.path.exists(version_file):
    version_data = regis.rex_json.load_file(version_file)           
    if version_data != None:
      return version_data["tag"]

  return None

def env_paths():
  envPath = os.environ["PATH"]
  paths = envPath.split(os.pathsep)
  return paths

def retrieve_header_filters(targetDir, projectName):
  filepath = os.path.join(targetDir, f"{projectName}.project")

  if not os.path.exists(filepath):
    raise Exception(f"clang tools project file doesn't exist, please check your sharpmake scripts: {filepath}")

  jsonBlob = regis.rex_json.load_file(filepath)
  return jsonBlob["HeaderFilters"]

def create_header_filter_regex(headerFilters : list[str]):
  res = ""

  for filter in headerFilters:
    res += "("
    res += filter
    res += ")|"

  res = res.removesuffix("|")
  return res

def find_file_in_folder(file, path : str):
  fileToFind = file.lower()
  subFilesOrFolders = os.listdir(path)
  for fileOrFolder in subFilesOrFolders:
    absPath = os.path.join(path, fileOrFolder)
    if os.path.isfile(absPath):
      file_name = Path(absPath).name.lower()
      if file_name == fileToFind:
        return absPath
  
  return ''

def find_file_in_paths(file, directories : list[str]):
  for path in directories:
    if not os.path.exists(path):
      continue

    result = find_file_in_folder(file, path)
    if result != '':
      return result

  return ''

def find_directory_in_paths(dir : str, directories : list[str]):
  dir = dir.replace('\\', '/')
  folders = dir.split('/')  
  num_folders = len(folders)

  for path in directories:
    path = path.replace('\\', '/')
    path_folders = path.split('/')

    dir_idx = -1

    if os.path.exists(os.path.join(path, dir)):
      return os.path.join(path, dir)

    dir_idx = num_folders - 1
    path_idx = len(path_folders) - 1
    if (len(path_folders) < num_folders):
      continue

    while dir_idx >= 0:
      dir_folder = folders[dir_idx]
      path_folder = path_folders[path_idx]
      
      if dir_folder != path_folder:
        break

      dir_idx -= 1
      path_idx -= 1

    if dir_idx == -1:
      if os.path.exists(path):
        return path
      else:
        regis.diagnostics.log_err(f"matching directory found, but doesn't exist: {path}")

  return None

def find_in_parent(path, toFind):
  curr_path = path

  while toFind not in os.listdir(curr_path):
    if Path(curr_path).parent == curr_path:
      regis.diagnostics.log_err(f"{toFind} not found in parents of {path}")
      return ''

    curr_path = Path(curr_path).parent

  return curr_path

def find_root():
  res = find_in_parent(os.getcwd(), "source")
  if (res == ''):
    regis.diagnostics.log_err(f"root not found")

  return res

def find_files_with_extension(path : str, extension : str):
  files = os.listdir(path)
  files_with_extension = []
  for file in files:
    if Path(file).suffix == extension:
      files_with_extension.append(file)

  return files_with_extension

def is_windows():
  return os.name == 'nt'

def run_subprocess(command):
  proc = subprocess.Popen(command)
  return proc

def run_and_get_output(command):
  proc = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
  output, errc = proc.communicate()

  return output.decode('utf-8'), proc.returncode

def run_subprocess_with_working_dir(command, workingDir):
  proc = subprocess.Popen(command, cwd=workingDir)
  return proc

def run_subprocess_with_callback(command, callback, filterLines):
  proc = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
  callback(proc.pid, proc.stdout, False, filterLines)
  callback(proc.pid, proc.stderr, True, filterLines)
  return proc

def wait_for_process(process):
  streamdata = process.communicate()[0]
  return process.returncode  

def is_executable(path):
  if is_windows():
    if Path(path).suffix == ".exe":
      return True
  else:
    return os.access(path, os.X_OK)

def find_all_files_in_folder(dir, toFindRegex):
  return list(Path(dir).rglob(toFindRegex))

def remove_folders_recursive(dir : str):
  if os.path.exists(dir):
    shutil.rmtree(dir)


# ============================================ 
#
# Author: Nick De Breuck
# Twitter: @nick_debreuck
# 
# File: test.py
# Copyright (c) Nick De Breuck 2023
#
# ============================================

import os
import threading
import time
import threading
import subprocess
import re
import regis.required_tools
import regis.util
import regis.task_raii_printing
import regis.rex_json
import regis.code_coverage
import regis.diagnostics
import regis.generation
import regis.build
import regis.dir_watcher

from pathlib import Path
from datetime import datetime

root_path = regis.util.find_root()
tool_paths_dict = regis.required_tools.tool_paths_dict
settings = regis.rex_json.load_file(os.path.join(root_path, regis.util.settingsPathFromRoot))
_pass_results = {}

iwyu_intermediate_dir = "iwyu"
clang_tidy_intermediate_dir = "clang_tidy"
unit_tests_intermediate_dir = "unit_tests"
coverage_intermediate_dir = "coverage"
asan_intermediate_dir = "asan"
ubsan_intermediate_dir = "ubsan"
fuzzy_intermediate_dir = "fuzzy"
auto_test_intermediate_dir = "auto_test"

def get_pass_results():
  return _pass_results

def _is_in_line(line : str, keywords : list[str]):
  regex = "((error).(cpp))|((error).(h))"

  for keyword in keywords:
    if keyword.lower() in line.lower():
      return not re.search(regex, line.lower()) # make sure that lines like 'error.cpp' don't return positive

  return False

def _symbolic_print(line, filterLines : bool = False):
  error_keywords = ["failed", "error"]
  warn_keywords = ["warning"]

  if _is_in_line(line, error_keywords):
    regis.diagnostics.log_err(line)
  elif _is_in_line(line, warn_keywords):
    regis.diagnostics.log_warn(line)
  elif not filterLines:
    regis.diagnostics.log_no_color(line)

def _default_output_callback(pid, output, isStdErr, filterLines):
  logs_dir = os.path.join(settings["intermediate_folder"], "logs")
  filename = f"output_{pid}.log"
  if isStdErr:
    filename = f"errors_{pid}.log"

  filepath = os.path.join(logs_dir, filename)
  if os.path.exists(filepath):
    os.remove(filepath)
  elif not os.path.exists(logs_dir):
    os.makedirs(logs_dir)

  with open(filepath, "a+") as f:

    for line in iter(output.readline, b''):
      new_line : str = line.decode('UTF-8')
      if new_line.endswith('\n'):
        new_line = new_line.removesuffix('\n')

      _symbolic_print(new_line, filterLines)      
      f.write(f"{new_line}\n")

    regis.diagnostics.log_info(f"full output saved to {filepath}")

def _run_include_what_you_use(fixIncludes = False, shouldClean : bool = True, singleThreaded : bool = False):
  """Run include what you use on the codebase"""
  def _run(iwyuPath, compdb, outputPath, impPath, lock):
      """Run the actual include-what-you-use command and save the output into a log file"""
      # create the command line to launch include-what-you-use
      cmd = ""
      cmd += f"py {iwyuPath} -v -p={compdb}"
      cmd += f" -- -Xiwyu --quoted_includes_first"
      
      if impPath != "" and os.path.exists(impPath):
        cmd += f" -Xiwyu --mapping_file={impPath}"

      # run include-what-you-use and save the output into a file
      output = subprocess.getoutput(cmd)
      with open(outputPath, "w") as f:
        f.write(output)
      output_lines = output.split('\n')

      # print the output using our color coding
      # to detect if it's an error, warning or regular log
      with lock:
        for line in output_lines:
          _symbolic_print(line)

        # log to the user that output has been saved
        regis.diagnostics.log_info(f"include what you use info saved to {outputPath}")
    
  task_print = regis.task_raii_printing.TaskRaiiPrint("running include-what-you-use")

  # to avoid clashing with other builds, we create our own intermediate folder
  intermediate_folder = os.path.join(root_path, settings["intermediate_folder"], settings["build_folder"], iwyu_intermediate_dir)

  # because we have the intermediates of iwyu in its own intermediate folder
  # we can safely clean it without affecting other builds
  if shouldClean:
    regis.diagnostics.log_info(f"cleaning {intermediate_folder}..")
    regis.util.remove_folders_recursive(intermediate_folder)

  # perform a new generation to make sure we actually have files to go over
  generated_files = _generate_test_files(regis.generation.create_config(f'-intermediate-dir={iwyu_intermediate_dir} -disable-clang-tidy-for-thirdparty -IDE None'))

  # get the compiler dbs that are just generated
  result = list(filter(lambda file: 'compile_commands.json' in file, generated_files))

  # find all the compiler dbs.
  # these act as input for include-what-you-use
  result = regis.util.find_all_files_in_folder(intermediate_folder, "compile_commands.json")
    
  threads : list[threading.Thread] = []
  output_files_per_project : dict[str, list] = {}
  lock = threading.Lock()

  iwyu_path = tool_paths_dict["include_what_you_use_path"]
  iwyu_tool_path = os.path.join(Path(iwyu_path).parent, "iwyu_tool.py")

  # create the include-what-you-use jobs
  for compiler_db in result:
    compiler_db_folder = Path(compiler_db).parent
    impPath = os.path.join(compiler_db_folder, "iwyu.imp")
    output_path = os.path.join(compiler_db_folder, "iwyu_output.log")
    project_name = _get_project_name_of_compdb(compiler_db_folder)

    # if we haven't compiled this project yet, prep the dict for it
    if project_name not in output_files_per_project:
      output_files_per_project[project_name] = []

    output_files_per_project[project_name].append(output_path)

    thread = threading.Thread(target=_run, args=(iwyu_tool_path, compiler_db, output_path, impPath, lock))
    thread.start()

    # very simple way of splitting single threaded with multi threaded mode
    # if we're single thread, we create a thread and immediately join it
    # if we're using multi threaded mode, we join them after all of them have been created
    if singleThreaded:
      thread.join() 
    else:
      threads.append(thread)

  # join all the threads after they've all been created
  # this is basically a no op if we're running in single threaded mode
  for thread in threads:
    thread.join()

  threads.clear()

  # because different configs could require different symbols or includes
  # we need to process all configs first, then process each output file for each config
  # for a given project and only if an include is not needed in all configs
  # take action and remove it or replace it with a forward declare
  # this can't be multithreaded
  if fixIncludes:
    regis.diagnostics.log_info(f'Applying fixes..')

  fix_includes_path = os.path.join(Path(iwyu_path).parent, "fix_includes.py")

  # this is the actual run in trying to fix the includes
  # however it can be faked when fixIncludes is false
  # if so, it'll do a dry run without changing anything
  # it'll still return a proper return code
  # indicating if anything needs to be changed
  rc = 0
  for key in output_files_per_project.keys():
    output_files = output_files_per_project[key]
    lines = []
    regis.diagnostics.log_info(f'processing: {key}')

    # include-what-you-use uses the output path of iwyu to determine what needs to be fixed
    # we merge all the outputs of all runs of iwyu on a project in different configs
    # and pass that temporary file over to include what you use
    for file in output_files:
      f = open(file, "r")
      lines.extend(f.readlines())

    filename = f'{key}_tmp.iwyu'
    filepath = os.path.join(intermediate_folder, filename)
    f = open(filepath, "w")
    f.writelines(lines)
    f.close()

    # create the fix includes command line
    cmd = f"py {fix_includes_path} --noreorder --process_merged=\"{filepath}\" --nocomments --nosafe_headers"

    if fixIncludes == False:
      cmd += f" --dry_run"

    # run the fix includes command line
    rc |= os.system(f"{cmd}")  

  return rc

# the compdbPath directory contains all the files needed to configure clang tools
# this includes the compiler database, clang tidy config files, clang format config files
# and a custom generated project file, which should have the same filename as the source root directory
# of the project you're testing
def _get_project_name_of_compdb(compdbPath):
  dirs = os.listdir(compdbPath)
  for dir in dirs:
    if ".project" in dir:
      return dir.split(".")[0]
  
  return ""

def _run_clang_tidy(filesRegex, shouldClean : bool = True, singleThreaded : bool = False, filterLines : bool = False, shouldFix : bool = False):
  """Run clang-tidy on the codebase"""
  rc = [0]
  def _run(cmd : str, rc : int):
    """Run the actual clang-tidy command"""
    regis.diagnostics.log_info(f"executing: {cmd}")
    proc = regis.util.run_subprocess_with_callback(cmd, _default_output_callback, filterLines)
    new_rc = regis.util.wait_for_process(proc)
    if new_rc != 0:
      regis.diagnostics.log_err(f"clang-tidy failed for {compiler_db}")
      regis.diagnostics.log_err(f"config file: {config_file_path}")
    rc[0] |= new_rc

  task_print = regis.task_raii_printing.TaskRaiiPrint("running clang-tidy")

  # to avoid clashing with other builds, we generate our projects and intermediate in our own folder
  intermediate_folder = os.path.join(root_path, settings["intermediate_folder"], settings["build_folder"], clang_tidy_intermediate_dir)

  # because our project and intermediates are generated into our own folder, this makes it safe to clean it
  if shouldClean:
    regis.diagnostics.log_info(f"cleaning {intermediate_folder}..")
    regis.util.remove_folders_recursive(intermediate_folder)

  # perform a new generation to make sure we actually have files to go over
  generated_files = _generate_test_files(regis.generation.create_config(f'-intermediate-dir={clang_tidy_intermediate_dir} -disable-clang-tidy-for-thirdparty -IDE None'))

  # get the compiler dbs that are just generated
  result = list(filter(lambda file: 'compile_commands.json' in file, generated_files))

  # create the clang-tidy jobs, we limit ourselves to 5 threads at the moment as running clang-tidy is quite performance heavy
  threads : list[threading.Thread] = []
  threads_to_use = 5
  script_path = os.path.dirname(__file__)
  clang_tidy_path = tool_paths_dict["clang_tidy_path"]
  clang_apply_replacements_path = tool_paths_dict["clang_apply_replacements_path"]

  for compiler_db in result:
    compiler_db_folder = Path(compiler_db).parent
    config_file_path = f"{compiler_db_folder}/.clang-tidy_second_pass"

    project_name = _get_project_name_of_compdb(compiler_db_folder)
    header_filters = regis.util.retrieve_header_filters(compiler_db_folder, project_name)
    header_filters_regex = regis.util.create_header_filter_regex(header_filters)
    
    # build up the clang-tidy command
    cmd = f"py \"{script_path}/run_clang_tidy.py\""
    cmd += f" -clang-tidy-binary=\"{clang_tidy_path}\""  # location of clang-tidy executable
    cmd += f" -clang-apply-replacements-binary=\"{clang_apply_replacements_path}\"" # location of clang-apply-replacements executable
    cmd += f" -config-file=\"{config_file_path}\"" # location of clang-tidy config file
    cmd += f" -p=\"{compiler_db_folder}\"" # location of compiler db folder (not the location to the file, but to its parent folder)
    cmd += f" -header-filter={header_filters_regex}" # only care about headers of the current project
    cmd += f" -quiet" # we don't want extensive logging
    cmd += f" -j={threads_to_use}" # only use a certain amount of threads, to reduce the performance overhead

    # auto fix found issues. This doesn't work for every enabled check.
    if shouldFix:
      cmd += f" -fix"

    # add the regex of the files we care about
    cmd += f" {filesRegex}"

    # perform an incremental run, avoid rescanning previous scanned files that didn't change (ignores cpp files if their headers changed)
    if not shouldClean:
      cmd += f" -incremental"

    # dirty hack to enable single thread mode vs multi threaded mode
    # in single threaded mode, we join the threads immediately
    thread = threading.Thread(target=_run, args=(cmd,rc,))
    thread.start()

    if singleThreaded:
      thread.join()
    else:
      threads.append(thread)

  for thread in threads:
    thread.join()

  return rc[0]

def _generate_test_files(config):
  """Perform a generation for a test"""
  settings_path = os.path.join(root_path, regis.util.settingsPathFromRoot)
  build_dir = os.path.join(settings["intermediate_folder"], settings["build_folder"])
  with regis.dir_watcher.DirWatcher(build_dir, True) as dir_watcher:
    proc = regis.generation.new_generation(settings_path, config)
    proc.wait()

  return dir_watcher.created_or_modified_files

def _build_files(configs : list[str], compilers : list[str], intermediateDir : str, projectsToBuild : list[str] = "", singleThreaded : bool = False):
  """Build certain projects under a intermediate directory in certain configs using certain compilers
  This is useful after a generation to make sure all projects are build
  """

  # This is the list that'll store the results of each build
  result_arr = []

  def _run(prj, cfg, comp, intermediateDir, shouldClean, result):
    """Launch a build and store its result"""
    result.append(regis.build.new_build(prj, cfg, comp, intermediateDir, shouldClean, slnFile="", buildDependencies=True))

  should_clean = False
  intermediate_folder = settings["intermediate_folder"]
  build_folder = settings["build_folder"]

  directory = os.path.join(root_path, intermediate_folder, build_folder, intermediateDir)

  threads : list[threading.Thread] = []

  # loop over the projects, configs and compilers and create a build for each combination
  for project in projectsToBuild:
    for config in configs:
      for compiler in compilers:
        thread = threading.Thread(target=_run, args=(project, config, compiler, directory, should_clean, result_arr))
        thread.start()

        # A dirty hack for singlethreaded mode
        # we always spawn a thread but in single threaded mode, we join it immediately
        if singleThreaded:
          thread.join()
        else:
          threads.append(thread)

  # in multi threaded mode, we join threads after all of them have spawned
  for thread in threads:
    thread.join()

  # if any result return code is different than 0
  # a build has failed somewhere
  return result_arr.count(0) != len(result_arr)

def _create_full_intermediate_dir(dir):
  """Create the absolute path for the test build directory"""
  return os.path.join(os.getcwd(), settings["intermediate_folder"], settings["build_folder"], dir)

# unit tests
def _generate_unit_tests(shouldClean):
  task_print = regis.task_raii_printing.TaskRaiiPrint("generating unit test projects")

  if shouldClean:
    # Clean the intermediates first if specified by the user
    # we clean in the generation step, to make sure that we only generate the unit tests we need
    full_intermediate_dir = _create_full_intermediate_dir(unit_tests_intermediate_dir)
    regis.diagnostics.log_info(f"cleaning {full_intermediate_dir}..")
    regis.util.remove_folders_recursive(full_intermediate_dir)

  # Generate the unit test projects and return what's generated
  return _generate_test_files(regis.generation.create_config(f'-intermediate-dir={clang_tidy_intermediate_dir} -disable-clang-tidy-for-thirdparty -no-clang-tools -enable-unit-tests'))

def _build_unit_tests(projects, singleThreaded : bool = False):
  """Build the unit tests. It's assumed the unit tests are already generated"""
  task_print = regis.task_raii_printing.TaskRaiiPrint("building tests")
  return _build_files(["debug", "debug_opt", "release"], ["msvc", "clang"], unit_tests_intermediate_dir, projects, singleThreaded)

def _run_unit_tests(unitTestPrograms):
  """Run all unit tests programs that got build in an earlier step"""
  task_print = regis.task_raii_printing.TaskRaiiPrint("running unit tests")
  
  rc = 0
  
  # loop over each unit test program path and run it
  for program in unitTestPrograms:
    regis.diagnostics.log_info(f"running: {Path(program).name}")
    proc = regis.util.run_subprocess(program)
    new_rc = regis.util.wait_for_process(proc)
    if new_rc != 0:
      regis.diagnostics.log_err(f"unit test failed for {program}") # use full path to avoid ambiguity
    rc |= new_rc

  return rc

# coverage
def _generate_coverage(shouldClean):
  """Generate the projects with coverage enabled"""
  task_print = regis.task_raii_printing.TaskRaiiPrint("generating coverage code")

  if shouldClean:
    # Clean all intermediates first if specified by the user.
    # we do this in the generation step so that we only generate coverage projects
    # for the projects that we care about
    full_intermediate_dir = _create_full_intermediate_dir(coverage_intermediate_dir)
    regis.diagnostics.log_info(f"cleaning {full_intermediate_dir}..")
    regis.util.remove_folders_recursive(full_intermediate_dir)

  return _generate_test_files(regis.generation.create_config(f'-intermediate-dir={coverage_intermediate_dir} -disable-clang-tidy-for-thirdparty -enable-code-coverage -enable-unit-tests'))

def _build_coverage(projects, singleThreaded : bool = False):
  """Build the projects with coverage enabled"""
  task_print = regis.task_raii_printing.TaskRaiiPrint("building coverage code")
  return _build_files(["coverage"], ["clang"], coverage_intermediate_dir, projects, singleThreaded)

def _run_coverage(unitTestPrograms):
  """Run the programs that have coverage enabled."""
  task_print = regis.task_raii_printing.TaskRaiiPrint("running coverage")

  raw_data_files = []

  rc = 0
  # LLVM uses the env variable LLVM_PROFILE_FILE
  # to hold the path where the coverage raw data should be saved to
  # running code coverage is very intensive
  # so we only allow 1 program to run at a time
  for program in unitTestPrograms:
    regis.diagnostics.log_info(f"running: {Path(program).name}")
    coverage_rawdata_filename = _get_coverage_rawdata_filename(program)
    output_path = os.path.join(Path(program).parent, coverage_rawdata_filename)
    raw_data_files.append(output_path)
    os.environ["LLVM_PROFILE_FILE"] = output_path # this is what llvm uses to set the raw data filename for the coverage data
    proc = regis.util.run_subprocess(program)
    new_rc = regis.util.wait_for_process(proc)
    if new_rc != 0:
      regis.diagnostics.log_err(f"unit test failed for {program}") # use full path to avoid ambiguity
    rc |= new_rc

  return raw_data_files

def _relocate_coverage_data(programsRun : list[str]):
  """Relocate the raw coverage data to a more appropriate location"""
  task_print = regis.task_raii_printing.TaskRaiiPrint("relocating coverage files")
  data_files = []

  for program in programsRun:
    coverage_rawdata_filename = _get_coverage_rawdata_filename(program)
    newPath = os.path.join(Path(program).parent, coverage_rawdata_filename)
    if (os.path.exists(newPath)):
      os.remove(newPath)
    os.rename(coverage_rawdata_filename, newPath)
    data_files.append(newPath)
    
  return data_files

def _index_rawdata_files(rawdataFiles : list[str]):
  task_print = regis.task_raii_printing.TaskRaiiPrint("indexing rawdata files")
  output_files = []

  for file in rawdataFiles:
    output_files.append(regis.code_coverage.create_index_rawdata(file))

  return output_files

def _create_coverage_reports(programsRun, indexdataFiles):
  task_print = regis.task_raii_printing.TaskRaiiPrint("creating coverage reports")

  rc = 0
  for index in range(len(programsRun)):
    program = programsRun[index]
    indexdata_file = indexdataFiles[index]

    if Path(program).stem != Path(indexdata_file).stem:
      rc = 1
      regis.diagnostics.log_err(f"program stem doesn't match coverage file stem: {Path(program).stem} != {Path(indexdata_file).stem}")

    regis.code_coverage.create_line_oriented_report(program, indexdata_file)
    regis.code_coverage.create_file_level_summary(program, indexdata_file)
    regis.code_coverage.create_lcov_report(program, indexdata_file)

  return rc

def _parse_coverage_reports(indexdataFiles):
  task_print = regis.task_raii_printing.TaskRaiiPrint("parsing coverage reports")

  rc = 0
  for indexdata_file in indexdataFiles:
    report_filename = regis.code_coverage.get_file_level_summary_filename(indexdata_file)
    rc |= regis.code_coverage.parse_file_summary(report_filename)

  return rc

def _get_coverage_rawdata_filename(program : str):
  return f"{Path(program).stem}.profraw"

# asan
def _generate_address_sanitizer(shouldClean):
  task_print = regis.task_raii_printing.TaskRaiiPrint("generating address sanitizer code")

  if shouldClean:
    full_intermediate_dir = _create_full_intermediate_dir(asan_intermediate_dir)
    regis.diagnostics.log_info(f"cleaning {full_intermediate_dir}..")
    regis.util.remove_folders_recursive(full_intermediate_dir)

  return _generate_test_files(regis.generation.create_config(f'-intermediate-dir={asan_intermediate_dir} -disable-clang-tidy-for-thirdparty -enable-unit-tests -enable-address-sanitizer -IDE None'))

def _build_address_sanitizer(projects, singleThreaded : bool = False):
  task_print = regis.task_raii_printing.TaskRaiiPrint("building address sanitizer code")
  return _build_files(["address_sanitizer"], ["clang"], asan_intermediate_dir, projects, singleThreaded)

def _run_address_sanitizer(unitTestPrograms):
  task_print = regis.task_raii_printing.TaskRaiiPrint("running address sanitizer tests")
  
  rc = 0
  for program in unitTestPrograms:
    regis.diagnostics.log_info(f"running: {Path(program).name}")
    log_folder_path = Path(program).parent
    log_folder = log_folder_path.as_posix()
    
    # for some reason, setting an absolute path for the log folder doesn't work
    # so we have to set the working directory of the program to where it's located so the log file will be there as well
    # ASAN_OPTIONS common flags: https://github.com/google/sanitizers/wiki/SanitizerCommonFlags
    # ASAN_OPTIONS flags: https://github.com/google/sanitizers/wiki/AddressSanitizerFlags
    asan_options = f"print_stacktrace=1:log_path=asan.log"
    os.environ["ASAN_OPTIONS"] = asan_options # print callstacks and save to log file
    
    proc = regis.util.run_subprocess_with_working_dir(program, log_folder)
    new_rc = regis.util.wait_for_process(proc)
    log_file_path = os.path.join(log_folder, f"asan.log.{proc.pid}")
    if new_rc != 0 or os.path.exists(log_file_path):
      regis.diagnostics.log_err(f"address sanitizer failed for {program}") # use full path to avoid ambiguity
      regis.diagnostics.log_err(f"for more info, please check: {log_file_path}")
      new_rc = 1
    rc |= new_rc

  return rc

# ubsan
def _generate_undefined_behavior_sanitizer(shouldClean):
  task_print = regis.task_raii_printing.TaskRaiiPrint("generating undefined behavior sanitizer code")

  if shouldClean:
    full_intermediate_dir = _create_full_intermediate_dir(ubsan_intermediate_dir)
    regis.diagnostics.log_info(f"cleaning {full_intermediate_dir}..")
    regis.util.remove_folders_recursive(full_intermediate_dir)

  return _generate_test_files(regis.generation.create_config(f'-intermediate-dir={ubsan_intermediate_dir} -disable-clang-tidy-for-thirdparty -enable-unit-tests -enable-ub-sanitizer -IDE None'))

def _build_undefined_behavior_sanitizer(projects, singleThreaded : bool = False):
  task_print = regis.task_raii_printing.TaskRaiiPrint("building undefined behavior sanitizer code")
  return _build_files(["undefined_behavior_sanitizer"], ["clang"], ubsan_intermediate_dir, projects, singleThreaded)

def _run_undefined_behavior_sanitizer(unitTestPrograms):
  task_print = regis.task_raii_printing.TaskRaiiPrint("running undefined behavior sanitizer tests")
  
  rc = 0
  for program in unitTestPrograms:
    regis.diagnostics.log_info(f"running: {Path(program).name}")
    log_folder_path = Path(program).parent
    log_folder = log_folder_path.as_posix()
    
    # for some reason, setting an absolute path for the log folder doesn't work
    # so we have to set the working directory of the program to where it's located so the log file will be there as well
    # UBSAN_OPTIONS common flags: https://github.com/google/sanitizers/wiki/SanitizerCommonFlags
    ubsan_options = f"print_stacktrace=1:log_path=ubsan.log"
    os.environ["UBSAN_OPTIONS"] = ubsan_options # print callstacks and save to log file
    proc = regis.util.run_subprocess_with_working_dir(program, log_folder)
    new_rc = regis.util.wait_for_process(proc)
    log_file_path = os.path.join(log_folder, f"ubsan.log.{proc.pid}")
    if new_rc != 0 or os.path.exists(log_file_path): # if there's a ubsan.log.pid created, the tool found issues
      regis.diagnostics.log_err(f"undefined behavior sanitizer failed for {program}") # use full path to avoid ambiguity
      regis.diagnostics.log_err(f"for more info, please check: {log_file_path}")
      new_rc = 1
    rc |= new_rc

  return rc

# fuzzy
def _generate_fuzzy_testing(shouldClean):
  task_print = regis.task_raii_printing.TaskRaiiPrint("generating fuzzy testing code")

  if shouldClean:
    full_intermediate_dir = _create_full_intermediate_dir(fuzzy_intermediate_dir)
    regis.diagnostics.log_info(f"cleaning {full_intermediate_dir}..")
    regis.util.remove_folders_recursive(full_intermediate_dir)

  return _generate_test_files(regis.generation.create_config(f"-enable-fuzzy-testing -intermediate-dir={fuzzy_intermediate_dir} -IDE None"))

def _build_fuzzy_testing(projects, singleThreaded : bool = False):
  task_print = regis.task_raii_printing.TaskRaiiPrint("building fuzzy testing code")
  return _build_files(["fuzzy"], ["clang"], fuzzy_intermediate_dir, projects, singleThreaded)

def _run_fuzzy_testing(fuzzyPrograms):
  task_print = regis.task_raii_printing.TaskRaiiPrint("running fuzzy tests")
  
  rc = 0
  for program in fuzzyPrograms:
    regis.diagnostics.log_info(f"running: {Path(program).name}")
    log_folder_path = Path(program).parent
    log_folder = log_folder_path.as_posix()
    
    # for some reason, setting an absolute path for the log folder doesn't work
    # so we have to set the working directory of the program to where it's located so the log file will be there as well
    # Can't use both ASAN as well as UBSAN options, so we'll set the same for both and hope that works
    # https://gcc.gnu.org/bugzilla/show_bug.cgi?id=94328
    # https://stackoverflow.com/questions/60774638/logging-control-for-address-sanitizer-plus-undefined-behavior-sanitizer
    asan_options = f"print_stacktrace=1:log_path=fuzzy.log"
    ubsan_options = f"print_stacktrace=1:log_path=fuzzy.log"
    os.environ["ASAN_OPTIONS"] = asan_options # print callstacks and save to log file
    os.environ["UBSAN_OPTIONS"] = ubsan_options # print callstacks and save to log file
    num_runs = 10000 # we'll run 10'000 fuzzy tests, should be more than enough
    proc = regis.util.run_subprocess_with_working_dir(f"{program} -runs={num_runs}", log_folder)
    new_rc = regis.util.wait_for_process(proc)
    log_file_path = os.path.join(log_folder, f"fuzzy.log.{proc.pid}")
    if new_rc != 0 or os.path.exists(log_file_path): # if there's a ubsan.log.pid created, the tool found issues
      regis.diagnostics.log_err(f"fuzzy testing failed for {program}") # use full path to avoid ambiguity
      if os.path.exists(log_file_path):
        regis.diagnostics.log_err(f"issues found while fuzzing!")
        regis.diagnostics.log_err(f"for more info, please check: {log_file_path}")
      new_rc = 1
    rc |= new_rc

  return rc

# auto tests
def _generate_auto_tests(shouldClean):
  task_print = regis.task_raii_printing.TaskRaiiPrint("generating auto tests")

  if shouldClean:
    full_intermediate_dir = _create_full_intermediate_dir(auto_test_intermediate_dir)
    regis.diagnostics.log_info(f"cleaning {full_intermediate_dir}..")
    regis.util.remove_folders_recursive(full_intermediate_dir)

  return _generate_test_files(regis.generation.create_config(f"-no-clang-tools -enable-auto-tests -intermediate-dir={auto_test_intermediate_dir} -IDE None"))

def _build_auto_tests(configs, compilers, projects, singleThreaded : bool = False):
  task_print = regis.task_raii_printing.TaskRaiiPrint("building auto tests")
  return _build_files(configs, compilers, auto_test_intermediate_dir, projects, singleThreaded)

def _process_tests_file(file, programs, timeoutInSeconds):
  json_blob = regis.rex_json.load_file(file)
    
  results = {}

  for test in json_blob:
    command_line = json_blob[test]["command_line"]

    rc = 0
    for program in programs:
      regis.diagnostics.log_info(f"running: {Path(program).name}")
      regis.diagnostics.log_info(f"with command line: {command_line}")
      proc = regis.util.run_subprocess(f"{program} {command_line}")

      # wait for program to finish on a different thread so we can terminate it on timeout
      thread = threading.Thread(target=lambda: proc.wait())
      thread.start()

      # wait for timeout to trigger or until the program exits
      now = time.time()
      duration = 0
      killed_process = False
      max_seconds = timeoutInSeconds
      while True:
        duration = time.time() - now
        if not thread.is_alive():
          break
        
        if duration > max_seconds:
          proc.terminate() 
          killed_process = True
          break

      # makes sure that we get an error code even if the program crashed
      proc.communicate()
      new_rc = proc.returncode
      
      if new_rc != 0:
        if killed_process:
          regis.diagnostics.log_err(f"auto test timeout triggered for {program} after {max_seconds} seconds") # use full path to avoid ambiguity
        else:
          rc |= new_rc
          regis.diagnostics.log_err(f"auto test failed for {program} with returncode {new_rc}") # use full path to avoid ambiguity

      results[program] = rc

  return results

def _find_files(folder, predicate):
  found_files : list[str] = []

  for root, dirs, files in os.walk(folder):
    for file in files:
      if predicate(file):
        path = os.path.join(root, file)
        found_files.append(path)      
  
  return found_files

def _run_auto_tests(programs, timeoutInSeconds):
  task_print = regis.task_raii_printing.TaskRaiiPrint("running auto tests")
  
  test_dir = os.path.join(root_path, settings["tests_folder"])

  files = _find_files(test_dir, lambda file: Path(file).name == "tests.json")

  results : list[dict] = []

  for file in files:
    results.append(_process_tests_file(file, programs, timeoutInSeconds))

  for res in results:
    values = list(res.values())
    if (len(values) != values.count(0)):
      return 1

  return 0

# public API
def test_include_what_you_use(shouldClean : bool = True, singleThreaded : bool = False, shouldFix : bool = False):
  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  rc = _run_include_what_you_use(shouldFix, shouldClean, singleThreaded)

  if rc != 0:
    regis.diagnostics.log_err(f"include-what-you-use pass failed")

  _pass_results["include-what-you-use"] = rc

def test_clang_tidy(filesRegex = ".*", shouldClean : bool = True, singleThreaded : bool = False, filterLines : bool = False, autoFix : bool = False):
  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  rc = _run_clang_tidy(filesRegex, shouldClean, singleThreaded, filterLines, autoFix)
  if rc != 0:
    regis.diagnostics.log_err(f"clang-tidy pass failed")

  _pass_results["clang-tidy"] = rc

def test_unit_tests(projects, shouldClean : bool = True, singleThreaded : bool = False):
  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  generated_files = _generate_unit_tests(shouldClean)
  rc = len(generated_files) == 0
  if rc != 0:
    regis.diagnostics.log_err(f"failed to generate tests")
  _pass_results["unit tests generation"] = rc

  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  with regis.dir_watcher.DirWatcher(_create_full_intermediate_dir(clang_tidy_intermediate_dir), True) as dir_watcher:
    rc |= _build_unit_tests(projects, singleThreaded)

  if rc != 0:
    regis.diagnostics.log_err(f"failed to build tests")
  _pass_results["unit tests building"] = rc

  executables = dir_watcher.filter_created_or_modified_files(lambda dir: dir.endswith('.exe'))

  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  rc |= _run_unit_tests(executables)
  if rc != 0:
    regis.diagnostics.log_err(f"unit tests failed")
  _pass_results["unit tests result"] = rc

def test_code_coverage(projects, shouldClean : bool = True, singleThreaded : bool = False):
  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  rc = _generate_coverage(shouldClean)
  if not rc:
    regis.diagnostics.log_err(f"failed to generate coverage")
  # _pass_results stores return codes
  # so we need store 0 for success, anything else for failure
  _pass_results["coverage generation"] = len(rc) == 0

  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  rc = _build_coverage(projects, singleThreaded)

  if rc != 0:
    regis.diagnostics.log_err(f"failed to build coverage")
  _pass_results["coverage building"] = rc

  executables = _find_files(_create_full_intermediate_dir(coverage_intermediate_dir), lambda file: file.endswith('.exe'))

  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  rawdata_files = _run_coverage(executables)
  
  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  indexdata_files = _index_rawdata_files(rawdata_files)
  
  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  rc |= _create_coverage_reports(executables, indexdata_files)
  if rc != 0:
    regis.diagnostics.log_err(f"failed to create coverage reports")
  _pass_results["coverage report creation"] = rc

  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  rc |= _parse_coverage_reports(indexdata_files)
  if rc != 0:
    regis.diagnostics.log_err(f"Not all the code was covered")
  _pass_results["coverage report result"] = rc

def test_asan(projects, shouldClean : bool = True, singleThreaded : bool = False):
  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  generated_files = _generate_address_sanitizer(shouldClean)
  rc = len(generated_files) == 0
  if rc != 0:
    regis.diagnostics.log_err(f"failed to generate asan code")
  _pass_results["address sanitizer generation"] = rc

  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  rc |= _build_address_sanitizer(projects, singleThreaded)

  if rc != 0:
    regis.diagnostics.log_err(f"failed to build asan code")
  _pass_results["address sanitizer building"] = rc
  
  executables = _find_files(_create_full_intermediate_dir(asan_intermediate_dir), lambda file: file.endswith('.exe'))

  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  rc |= _run_address_sanitizer(executables)
  if rc != 0:
    regis.diagnostics.log_err(f"invalid code found with asan")
  _pass_results["address sanitizer result"] = rc

def test_ubsan(projects, shouldClean : bool = True, singleThreaded : bool = False):
  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  generated_files = _generate_undefined_behavior_sanitizer(shouldClean)
  rc = len(generated_files) == 0
  if rc != 0:
    regis.diagnostics.log_err(f"failed to generate ubsan code")
  _pass_results["undefined behavior sanitizer generation"] = rc
  
  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  with regis.dir_watcher.DirWatcher('.', True) as dir_watcher:
    rc |= _build_undefined_behavior_sanitizer(projects, singleThreaded)

  if rc != 0:
    regis.diagnostics.log_err(f"failed to build ubsan code")
  _pass_results["undefined behavior sanitizer building"] = rc
  
  executables = _find_files(_create_full_intermediate_dir(ubsan_intermediate_dir), lambda file: file.endswith('.exe'))

  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  rc |= _run_undefined_behavior_sanitizer(executables)
  if rc != 0:
    regis.diagnostics.log_err(f"invalid code found with ubsan")
  _pass_results["undefined behavior sanitizer result"] = rc

def test_fuzzy_testing(projects, shouldClean : bool = True, singleThreaded : bool = False):
  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  generated_files = _generate_fuzzy_testing(shouldClean)
  rc = len(generated_files) == 0

  if rc != 0:
    regis.diagnostics.log_err(f"failed to generate fuzzy code")
  _pass_results["fuzzy testing generation"] = rc
  
  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  with regis.dir_watcher.DirWatcher('.', True) as dir_watcher:
    rc |= _build_fuzzy_testing(projects, singleThreaded)

  if rc != 0:
    regis.diagnostics.log_err(f"failed to build fuzzy code")
  _pass_results["fuzzy testing building"] = rc

  executables = _find_files(_create_full_intermediate_dir(fuzzy_intermediate_dir), lambda file: file.endswith('.exe'))

  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  rc |= _run_fuzzy_testing(executables)
  if rc != 0:
    regis.diagnostics.log_err(f"invalid code found with fuzzy")
  _pass_results["fuzzy testing result"] = rc

def run_auto_tests(configs, compilers, projects, timeoutInSeconds : int, shouldClean : bool = True, singleThreaded : bool = False):
  rc = 0

  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  generated_files = _generate_auto_tests(shouldClean)
  rc = len(generated_files) == 0
  
  if rc != 0:
    regis.diagnostics.log_err(f"failed to generate auto test code")
  _pass_results["auto testing generation"] = rc
  
  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  with regis.dir_watcher.DirWatcher('.', True) as dir_watcher:
    rc |= _build_auto_tests(configs, compilers, projects, singleThreaded)

  if rc != 0:
    regis.diagnostics.log_err(f"failed to build auto test code")
  _pass_results["auto testing building"] = rc
  
  executables = _find_files(_create_full_intermediate_dir(auto_test_intermediate_dir), lambda file: file.endswith('.exe'))

  regis.diagnostics.log_no_color("-----------------------------------------------------------------------------")
  rc |= _run_auto_tests(executables, timeoutInSeconds)
  if rc != 0:
    regis.diagnostics.log_err(f"auto tests failed")
  _pass_results["auto testing result"] = rc



import time
import regis.diagnostics

class TaskRaiiPrint(object):
  def __init__(self, msg):
    self._msg = msg
    self._finished_msg = "done"
    self.start_time = time.time()

    regis.diagnostics.log_info(msg)
  
  def failed(self):
    self._finished_msg = "failed"

  def __del__(self):
    end_time = time.time()
    regis.diagnostics.log_info(f"{self._msg} - {self._finished_msg}")
    regis.diagnostics.log_info(f"  took {end_time - self.start_time:0.2f} seconds")


import os
import shutil
import regis.util
import regis.task_raii_printing

def __install(hooksPath):
    root_path = regis.util.find_root()

    if not os.path.exists(hooksPath):
        return
    
    hooks = os.listdir(hooksPath)

    for hook in hooks:
        src = os.path.join(hooksPath, hook)
        dst = os.path.join(root_path, ".git", "hooks", hook)
        shutil.copy(src, dst)

def run(hooksPath):
    task = regis.task_raii_printing.TaskRaiiPrint("Installing git hooks")
    __install(hooksPath)


import subprocess
import json
import regis.diagnostics

class _VisualStudioCatalog():
  def __init__(self, jsonData : dict):
    self.build_branch : str = jsonData['buildBranch']
    self.build_version : str = jsonData['buildVersion']
    self.id : str = jsonData['id']
    self.local_build : str = jsonData['localBuild']
    self.manifest_name : str = jsonData['manifestName']
    self.manifest_type : str = jsonData['manifestType']
    self.product_display_version : str = jsonData['productDisplayVersion']
    self.product_line : str = jsonData['productLine']
    self.product_line_version : str = jsonData['productLineVersion']
    self.product_milestone : str = jsonData['productMilestone']
    self.product_milestone_is_pre_release : str = jsonData['productMilestoneIsPreRelease']
    self.product_name : str = jsonData['productName']
    self.product_patch_version : str = jsonData['productPatchVersion']
    self.product_pre_release_milestone_suffix : str = jsonData['productPreReleaseMilestoneSuffix']
    self.product_semantic_version : str = jsonData['productSemanticVersion']
    self.required_engine_version : str = jsonData['requiredEngineVersion']

class _VisualStudioProperties():
  def __init__(self, jsonData : dict):
    self.campaign_id = jsonData['campaignId']
    self.channel_manifest_id = jsonData['channelManifestId']
    self.nickname = jsonData['nickname']
    self.setup_engine_filepath = jsonData['setupEngineFilePath']

# JSON data is loaded as output from the vswhere.exe program
# which comes from a visual studio install
class VisualStudioInstall():
  def __init__(self, jsonData):
    self.instance_id : str = jsonData['instanceId']
    self.install_date : str = jsonData['installDate']
    self.installation_name : str = jsonData['installationName']
    self.installation_path : str = jsonData['installationPath']
    self.installation_version : str = jsonData['installationVersion']
    self.product_id : str = jsonData['productId']
    self.product_path : str = jsonData['productPath']
    self.state : int = jsonData['state']
    self.is_complete : bool = jsonData['isComplete']
    self.is_launchable : bool = jsonData['isLaunchable']
    self.is_prerelease : bool = jsonData['isPrerelease']
    self.is_reboot_required : bool = jsonData['isRebootRequired']
    self.display_name : str = jsonData['displayName']
    self.description : str = jsonData['description']
    self.channel_id : str = jsonData['channelId']
    self.channel_url : str = jsonData['channelUri']
    self.engine_path : str = jsonData['enginePath']
    self.installed_channel_id : str = jsonData['installedChannelId']
    self.installed_channel_url : str = jsonData['installedChannelUri']
    self.release_notes : str = jsonData['releaseNotes']
    self.third_party_notices : str = jsonData['thirdPartyNotices']
    self.update_date : str = jsonData['updateDate']
    self.catalag = _VisualStudioCatalog(jsonData['catalog'])
    self.properties = _VisualStudioProperties(jsonData['properties'])

def installed_versions():
  vswhere_path = r"C:\Program Files (x86)\Microsoft Visual Studio\Installer\vswhere.exe"
  output = subprocess.getoutput(f'"{vswhere_path}" -format json')
  installed_versions : list[VisualStudioInstall] = []

  try:
    jsondata = json.loads(output)
    for installed_version in jsondata:
      installed_versions.append(VisualStudioInstall(installed_version))
  except Exception as ex:
    regis.diagnostics.log_warn(f'Failed to load visual studio versions.')
    regis.diagnostics.log_warn(f'Json data could not be loaded from vswhere.exe')
    regis.diagnostics.log_warn(f'Exception: {ex}')

  return installed_versions



import json
import os
import regis.diagnostics

def load_file(path):
  if not os.path.exists(path):
    regis.diagnostics.log_err(f"Failed to load json, file does not exist '{path}'")
    return None

  f = open(path)
  data = json.load(f)
  return data

def save_file(path : str, data):
  f = open(path, "w")
  json.dump(data, f, indent=2)


import os
import copy
import argparse
import shlex
import regis.diagnostics
import regis.rex_json
import regis.util
import regis.required_tools
import regis.subproc
import regis.diagnostics

from pathlib import Path

root = regis.util.find_root()
settings = regis.rex_json.load_file(os.path.join(root, regis.util.settingsPathFromRoot))
temp_dir = os.path.join(root, settings["intermediate_folder"])
tools_install_dir = os.path.join(temp_dir, settings["tools_folder"])
tool_paths_filepath = os.path.join(tools_install_dir, "tool_paths.json")
tool_paths_dict = regis.rex_json.load_file(tool_paths_filepath)

def _find_sharpmake_files(directory):
  sharpmakes_files = []
  for root, dirs, files in os.walk(directory):
    for file in files:
      extensions = Path(file).suffixes
      if len(extensions) == 2:
        if extensions[0] == ".sharpmake" and extensions[1] == ".cs":
          path = os.path.join(root, file)
          sharpmakes_files.append(path)
  
  return sharpmakes_files

def _find_sharpmake_root_files(directory):
  sharpmakes_files = []
  for root, dirs, files in os.walk(directory):
    for file in files:
      extensions = Path(file).suffixes
      if len(extensions) == 1:
        if extensions[0] == ".cs":
          path = os.path.join(root, file)
          sharpmakes_files.append(path)

  return sharpmakes_files

def _scan_for_sharpmake_files(settings : dict):
  """
  scans for sharpmake files in the current directory using the settings.
  it searches for all the sharpmake files in the sharpmake root, source folder and test folder.
  all searches are done recursively.
  """
  sharpmake_root = os.path.join(root, "_build", "sharpmake")
  source_root = os.path.join(root, settings["source_folder"])
  tests_root = os.path.join(root, settings["tests_folder"])
  
  sharpmakes_files = []
  sharpmakes_files.extend(_find_sharpmake_root_files(sharpmake_root))
  sharpmakes_files.extend(_find_sharpmake_files(source_root))
  sharpmakes_files.extend(_find_sharpmake_files(tests_root))

  return sharpmakes_files

def _save_config_file(settings : dict, config : dict):
  """Create a new config file. This file will be passed over to sharpmake"""

  config_dir = os.path.join(os.path.join(root, settings['intermediate_folder'], settings['build_folder']))
  if not os.path.exists(config_dir):
    os.mkdir(config_dir)

  config_path = os.path.join(config_dir, 'config.json')
  regis.rex_json.save_file(config_path, config)

  return config_path.replace('\\', '/')

def _make_optional_arg(arg : str):
  return f'-{arg}'

def _add_config_arguments(parser : argparse.ArgumentParser, defaultConfig : dict):
  """Load the sharpmake config file from disk and add the options as arguments to this script."""
  
  for setting in defaultConfig:
    arg = _make_optional_arg(setting)
    val = defaultConfig[setting]['Value']
    desc = defaultConfig[setting]['Description']
    if 'Options' in defaultConfig[setting]:
      desc += f' Options: {defaultConfig[setting]["Options"]}'
    
    if type(val) == bool:
      parser.add_argument(arg, help=desc, action='store_true')
    else:
      parser.add_argument(arg, help=desc, default=val)

def _load_default_config():
  return regis.rex_json.load_file(os.path.join(root, "_build", "sharpmake", "data", "default_config.json"))

def add_config_arguments_to_parser(parser):
  _add_config_arguments(parser, _load_default_config())

def create_config(args):
  """Create a config dictionary based on the arguments passed in."""

  if type(args) == str:
    parser = argparse.ArgumentParser()
    add_config_arguments_to_parser(parser)
    args = parser.parse_args(shlex.split(args))

  default_config : dict = regis.rex_json.load_file(os.path.join(root, "_build", "sharpmake", "data", "default_config.json"))
  config : dict = copy.deepcopy(default_config)
  for arg in vars(args):
    arg_name = arg
    arg_name = arg_name.replace('_', '-') # python converts all hyphens into underscores so we need to convert them back
    arg_val = getattr(args, arg)

    if arg_name in config:
      config[arg_name]['Value'] = arg_val

  return config

def new_generation(settingsPath : str, config : dict):
  """
  performs a new generation using the sharpmake files found by searching the current directory recursively.
  '/diagnostics' is always added as a sharpmake arguments.
  """

  # save the config file to disk
  settings = regis.rex_json.load_file(settingsPath)
  config_path = _save_config_file(settings, config)

  # scan recursively to find all the sharpmake files
  sharpmake_files = _scan_for_sharpmake_files(settings)
  
  # load the path where the sharpmake executable is located
  sharpmake_path = tool_paths_dict["sharpmake_path"]
  if len(sharpmake_path) == 0:
    regis.diagnostics.log_err("Failed to find sharpmake path")
    return

  # make sure the sharpmake files are quoted, that's expected by sharpmake
  sharpmake_sources = ""
  for sharpmake_file in sharpmake_files:
    sharpmake_sources += "\""
    sharpmake_sources += sharpmake_file
    sharpmake_sources += "\", "

  # replace all backwards slashes by forward slashes
  sharpmake_sources = sharpmake_sources[0:len(sharpmake_sources) - 2]
  sharpmake_sources = sharpmake_sources.replace('\\', '/')

  # run the actual executable
  return regis.subproc.run(f"{sharpmake_path} /sources({sharpmake_sources}) /diagnostics /configFile(\"{config_path}\")")
